import os
import sys
import traceback
from pathlib import Path
from typing import Type

from sssimp import APP_DIR


def mkdir(path):
    """
    Creates a directory or the parent directory if `path` is not a directory.
    Safely ignores if already exists
    """
    path = Path(path)
    if not path.is_dir():
        path = path.parent
    path.mkdir(exist_ok=True, parents=True)


def path_strip(path, parent=APP_DIR):
    """Strips `parent` from `path`"""
    path = str(path)
    # TODO 3.9+:
    # path = path.removeprefix(str(parent))
    # path = path.removeprefix(os.path.sep)
    # 3.8 compat:
    if path.startswith(str(parent)):
        path = path[len(str(parent)) :]
    if path.startswith(os.path.sep):
        path = path[1:]
    return path


def run_safely(f, exception_type: Type[Exception] = Exception, out=sys.stderr):
    try:
        f()
    except exception_type:
        print(traceback.format_exc(), file=out)



import json
import logging
from collections import namedtuple
from datetime import datetime
from io import StringIO

import pymdownx.emoji
from jinja2 import TemplateNotFound
from markdown import Extension, Markdown
from markdown.inlinepatterns import SimpleTagInlineProcessor

import sssimp
from sssimp import jinja
from sssimp.generators.html import PAGES, Page
from sssimp.utils import path_strip

# Set this if we want to have a fixed version for twemoji
# I'm considering this to have reproducible tests, but I'm not sure if it's a
# desirable implementation, since it might cause broken images in the future.
# pymdownx.emoji.TWEMOJI_PNG_CDN = (
#     "https://cdn.jsdelivr.net/gh/jdecked/twemoji@latest/assets/72x72/"
# )


class MarkdownPage(Page):
    def __init__(self, src, target):
        super().__init__(src, target)
        with open(self.src) as f:
            self.content = f.read()
        self.vars["plain_text"] = markdown_to_text(self.content)
        result = markdown_to_html(self.content)
        self.meta = {
            key: (
                json.loads(value[-1][1:])
                if value[-1].startswith("=")
                else value[-1]
            )
            for key, value in result.meta.items()
        }
        self.meta.setdefault("template", f"{self.src.parent.name}.html")
        self.vars["meta"] = self.meta
        self.vars["markdown"] = result.html

    def get_template(self):
        return jinja.get_template(self.meta["template"])

    @property
    def created_at(self):
        if "date" in self.meta:
            return datetime.fromisoformat(self.meta["date"])
        return super().created_at

    @property
    def title(self):
        if "title" in self.meta:
            return self.meta["title"]
        result = self.target.stem
        for char in "-_":
            result = result.replace(char, " ")
        return result.capitalize()


class StrikeSubExtension(Extension):
    """
    Strike text with `~` and underline text with `_`.
    """

    @staticmethod
    def _add_pattern(md, char, tag):
        proc = SimpleTagInlineProcessor(
            r"(\{0}\{0})(.+?)(\{0}\{0})".format(char), tag
        )
        md.inlinePatterns.register(proc, tag, 200)

    def extendMarkdown(self, md):
        self._add_pattern(md, "~", "del")
        self._add_pattern(md, "_", "ins")


def unmark_element(element, stream=None):
    if stream is None:
        stream = StringIO()
    if element.text:
        stream.write(element.text)
    for sub in element:
        unmark_element(sub, stream)
    if element.tail:
        stream.write(element.tail)
    return stream.getvalue()


Markdown.output_formats["plain"] = unmark_element


def markdown_to_html(text):
    HtmlWithMeta = namedtuple("HtmlWithMeta", ("html", "meta"))
    markdown = Markdown(
        extensions=[
            # Builtin extensions
            # cf https://python-markdown.github.io/extensions/ (in order)
            "abbr",
            "fenced_code",
            "footnotes",
            "admonition",
            "codehilite",
            "meta",
            "smarty",
            "toc",
            # Custom extensions
            StrikeSubExtension(),
            # Third-party extensions
            "pymdownx.emoji",
            "pymdownx.magiclink",
        ],
        extension_configs={
            "smarty": {"smart_angled_quotes": True},
            "pymdownx.emoji": {
                "alt": "html_entity",
                "emoji_index": pymdownx.emoji.twemoji,
                "options": {"classes": "emoji"},
            },
        },
        output_format="html",
        tab_length=2,
    )
    return HtmlWithMeta(markdown.convert(text), markdown.Meta)


def markdown_to_text(text):
    markdown = Markdown(
        extensions=["abbr", "meta", "admonition", StrikeSubExtension()],
        output_format="plain",
    )
    markdown.stripTopLevelTags = False
    result = markdown.convert(text)
    result = result.replace("[TOC]", "")
    return result


def prepare():
    for file in sssimp.CONTENT_DIR.glob("**/*.md"):
        as_html = path_strip(file.with_suffix(".html"), sssimp.CONTENT_DIR)
        target = sssimp.OUTPUT_DIR / as_html
        page = MarkdownPage(src=file, target=target)
        sssimp.IGNORE_ASSETS.add(str(file))
        try:
            page.get_template()
        except TemplateNotFound:
            logging.info(f"No matching template for {page}, ignoring file")
            continue
        PAGES.add(page)



import functools
import importlib.metadata
from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from . import config

__pkg__ = __name__.split(".")[-1]
__version__ = importlib.metadata.version(__pkg__)
print(f"sssimp version {__version__}")

APP_DIR = Path(__file__).parent
INPUT_DIR = Path(config.INPUT_PATH)
OUTPUT_DIR = Path(config.OUTPUT_PATH)
IGNORE_ASSETS = set()

CONTENT_DIR = None

jinja = Environment(
    autoescape=config.AUTOESCAPE,
    trim_blocks=True,
    lstrip_blocks=True,
    extensions=["jinja2.ext.do"],
)


def resolve():
    global CONTENT_DIR
    CONTENT_DIR = INPUT_DIR / config.CONTENT_DIR_NAME
    loader = FileSystemLoader(INPUT_DIR / "templates", followlinks=True)
    jinja.loader = loader


def filter(f):
    jinja.filters[f.__name__] = f

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


from . import filters



import json
import logging
import os
from pathlib import Path

import yaml

import sssimp
from sssimp import jinja
from sssimp.utils import path_strip

DATA_DIR = sssimp.INPUT_DIR / "data"


class Data:
    def __init__(self):
        self.dict = {}

    def handle_file(self, path: Path):
        parents = [
            x.split(".")[0] for x in path_strip(path, DATA_DIR).split(os.path.sep)
        ]
        getter = {
            "yml": yaml.safe_load,
            "yaml": yaml.safe_load,
            "json": json.load,
        }.get(path.suffix[1:])
        if not getter:
            return
        with open(path) as f:
            data = getter(f)
        target = self.dict
        for parent in parents[:-1]:
            target = target.setdefault(parent, {})
        target[parents[-1]] = data


def prepare():
    data = Data()
    for file in DATA_DIR.rglob("*.*"):
        logging.info(f"Handling data file {file}")
        data.handle_file(file)
    jinja.globals["data"] = data.dict



import importlib
import logging
import os
import shutil
import time
from argparse import ArgumentParser
from pathlib import Path

import sssimp
from sssimp import config
from sssimp.utils import mkdir, path_strip, run_safely

logging.basicConfig(level=config.LOG_LEVEL, format=config.LOG_FORMAT)


def run_generators():
    modules = set()
    for file in (sssimp.APP_DIR / "generators").glob("**/*.py"):
        module_str = path_strip(file).replace(os.path.sep, ".")[: -len(".py")]
        module = importlib.import_module(f".{module_str}", package="sssimp")
        if hasattr(module, "main"):
            modules.add(module)
        if hasattr(module, "prepare"):
            logging.info(f"Preparing {module.__name__}")
            run_safely(module.prepare)
    for module in modules:
        logging.info(f"Running {module.__name__}")
        run_safely(module.main)


def is_ignored(path: Path):
    for ignore_path in sssimp.IGNORE_ASSETS:
        if ignore_path == str(path):
            return ignore_path
        if Path(ignore_path).is_dir() and ignore_path in {
            str(x) for x in path.parents
        }:
            return ignore_path


def copy_assets():
    logging.debug(f"Ignore list: {sssimp.IGNORE_ASSETS}")
    for file in sssimp.CONTENT_DIR.glob("**/*"):
        if not file.is_file():
            continue
        relative_path = path_strip(file, sssimp.CONTENT_DIR)
        if match := is_ignored(file):
            logging.info(
                f"Ignoring asset {relative_path} because it matches"
                f' "{match}" from ignore list'
            )
            continue
        out_file = sssimp.OUTPUT_DIR / relative_path
        mkdir(out_file)
        logging.info(f"Copying raw asset {file} -> {out_file}")
        shutil.copyfile(file, out_file)


parser = ArgumentParser()
parser.add_argument("output_dir", nargs="?", default="output", type=Path)
parser.add_argument("--input", dest="input_dir", default="input", type=Path)
parser.add_argument(
    "--watch",
    action="store_true",
    help="Wait for changes and rebuild as soon as one occurs",
)


def run():
    if config.CLEAN_OUTPUT and sssimp.OUTPUT_DIR.exists():
        try:
            shutil.rmtree(sssimp.OUTPUT_DIR)
        except Exception:
            for f in sssimp.OUTPUT_DIR.glob("*"):
                shutil.rmtree(f)
        logging.info(f"Deleted {sssimp.OUTPUT_DIR}")
    logging.info("Running generators")
    run_generators()
    logging.info("Copying raw assets")
    copy_assets()
    print()
    print(
        f"Generation finished, open {sssimp.OUTPUT_DIR.resolve().as_uri()} in browser"
    )


def get_latest_ctime():
    return max(file.stat().st_ctime for file in sssimp.INPUT_DIR.rglob("*"))


def wait_for_change():
    print("Waiting for changes...")
    current = get_latest_ctime()
    while current == get_latest_ctime():
        time.sleep(config.WATCH_WAIT_TIME)


def main():
    args = parser.parse_args()
    sssimp.OUTPUT_DIR = args.output_dir
    sssimp.INPUT_DIR = args.input_dir
    sssimp.resolve()
    while True:
        run()
        if not args.watch:
            break
        try:
            wait_for_change()
        except KeyboardInterrupt:
            print("Stopped watcher")
            break


if __name__ == "__main__":
    main()
